/* math - Elementwise arithmetic functions                             */
/* XLISP-STAT 2.1 Copyright (c) 1990, by Luke Tierney                  */
/* Additions to Xlisp 2.1, Copyright (c) 1989 by David Michael Betz    */
/* You may give out copies of this software; for conditions see the    */
/* file COPYING included with this distribution.                       */
 
#include "xlisp.h"
#include "xlstat.h"

/*************************************************************************/
/*************************************************************************/
/**                                                                     **/
/**                 Recursive Vectorized Math Functions                 **/
/**                                                                     **/
/*************************************************************************/
/*************************************************************************/

/* The basic math functions have been modified to operate element-wise   */
/* on compound data. The operation is recursive: if compound data items  */
/* contain compound data items the mapping proceeds down to the next     */
/* level.                                                                */

LVAL xsradd(V)   { return (recursive_subr_map_elements(xadd, xsradd)); }
LVAL xsrsub(V)   { return (recursive_subr_map_elements(xsub, xsrsub)); }
LVAL xsrmul(V)   { return (recursive_subr_map_elements(xmul, xsrmul)); }
LVAL xsrdiv(V)   { return (recursive_subr_map_elements(xdiv, xsrdiv)); }
LVAL xsrmin(V)   { return (recursive_subr_map_elements(xmin, xsrmin)); }
LVAL xsrmax(V)   { return (recursive_subr_map_elements(xmax, xsrmax)); }
LVAL xsrrem(V)   { return (recursive_subr_map_elements(xrem, xsrrem)); }
LVAL xsrmod(V)   { return (recursive_subr_map_elements(xmod, xsrmod)); }
LVAL xsrexpt(V)  { return (recursive_subr_map_elements(xexpt, xsrexpt));}
LVAL xsrlog(V)   { return (recursive_subr_map_elements(xlog, xsrlog)); }

LVAL xsrlogand(V)   { return (recursive_subr_map_elements(xlogand, xsrlogand)); }
LVAL xsrlogior(V)   { return (recursive_subr_map_elements(xlogior, xsrlogior)); }
LVAL xsrlogxor(V)   { return (recursive_subr_map_elements(xlogxor, xsrlogxor)); }
LVAL xsrlognot(V)   { return (recursive_subr_map_elements(xlognot, xsrlognot)); }

LVAL xsrabs(V)   { return (recursive_subr_map_elements(xabs, xsrabs)); }
LVAL xsradd1(V)  { return (recursive_subr_map_elements(xadd1, xsradd1)); }
LVAL xsrsub1(V)  { return (recursive_subr_map_elements(xsub1, xsrsub1)); }
LVAL xsrsin(V)   { return (recursive_subr_map_elements(xsin, xsrsin)); }
LVAL xsrcos(V)   { return (recursive_subr_map_elements(xcos, xsrcos)); }
LVAL xsrtan(V)   { return (recursive_subr_map_elements(xtan, xsrtan)); }
LVAL xsrexp(V)   { return (recursive_subr_map_elements(xexp, xsrexp)); }
LVAL xsrsqrt(V)  { return (recursive_subr_map_elements(xsqrt, xsrsqrt)); }
LVAL xsrfloat(V) { return (recursive_subr_map_elements(xfloat, xsrfloat)); }
LVAL xsrrand(V)  { return (recursive_subr_map_elements(xrand, xsrrand)); }
LVAL xsrasin(V)  { return (recursive_subr_map_elements(xasin, xsrasin)); }
LVAL xsracos(V)  { return (recursive_subr_map_elements(xacos, xsracos)); }
LVAL xsratan(V)  { return (recursive_subr_map_elements(xatan, xsratan)); }
LVAL xsrphase(V) { return (recursive_subr_map_elements(xphase, xsrphase)); }

LVAL xsrfloor(V) { return(recursive_subr_map_elements(xfloor, xsrfloor)); }
LVAL xsrceil(V)  { return(recursive_subr_map_elements(xceil, xsrceil)); }
LVAL xsrfix(V)   { return(recursive_subr_map_elements(xfix, xsrfix)); }
LVAL xsrround(V) { return(recursive_subr_map_elements(xround, xsrround)); }

LVAL xsrminusp(V){ return (recursive_subr_map_elements(xminusp, xsrminusp)); }
LVAL xsrzerop(V) { return (recursive_subr_map_elements(xzerop, xsrzerop)); }
LVAL xsrplusp(V) { return (recursive_subr_map_elements(xplusp, xsrplusp)); }
LVAL xsrevenp(V) { return (recursive_subr_map_elements(xevenp, xsrevenp)); }
LVAL xsroddp(V)  { return (recursive_subr_map_elements(xoddp, xsroddp)); }

LVAL xsrlss(V){ return (recursive_subr_map_elements(xlss, xsrlss)); }
LVAL xsrleq(V){ return (recursive_subr_map_elements(xleq, xsrleq)); }
LVAL xsrequ(V){ return (recursive_subr_map_elements(xequ, xsrequ)); }
LVAL xsrneq(V){ return (recursive_subr_map_elements(xneq, xsrneq)); }
LVAL xsrgeq(V){ return (recursive_subr_map_elements(xgeq, xsrgeq)); }
LVAL xsrgtr(V){ return (recursive_subr_map_elements(xgtr, xsrgtr)); }

